"""aiohttp_oauthlib client.

Adapted from: https://gist.github.com/kellerza/5ca798f49983bb702bc6e7a05ba53def

- Fixed some implementation errors of the original (e.g. missing awaits)
- Allow modifier function on fetch_token and refresh_token in order to allow
  JSON Content-Type.
"""
import logging
import aiohttp

from oauthlib.common import generate_token, urldecode
from oauthlib.oauth2 import WebApplicationClient, InsecureTransportError
from oauthlib.oauth2 import TokenExpiredError, is_secure_transport


_LOGGER = logging.getLogger(__name__)


class TokenUpdated(Warning):
    """Exception."""
    def __init__(self, token):
        super().__init__()
        self.token = token


class OAuth2Session(aiohttp.ClientSession):
    """Versatile OAuth 2 extension to :class:`aiohttp.ClientSession`.

    Supports any grant type adhering to :class:`oauthlib.oauth2.Client` spec
    including the four core OAuth 2 grants.

    Can be used to create authorization urls, fetch tokens and access protected
    resources using the :class:`requests.Session` interface you are used to.

    - :class:`oauthlib.oauth2.WebApplicationClient` (default):
        Authorization Code Grant
    - :class:`oauthlib.oauth2.MobileApplicationClient`: Implicit Grant
    - :class:`oauthlib.oauth2.LegacyApplicationClient`:
        Password Credentials Grant
    - :class:`oauthlib.oauth2.BackendApplicationClient`:
        Client Credentials Grant

    Note that the only time you will be using Implicit Grant from python is if
    you are driving a user agent able to obtain URL fragments.
    """

    def __init__(
            self, client_id=None, client=None, auto_refresh_url=None,
            auto_refresh_kwargs=None, scope=None, redirect_uri=None,
            token=None, state=None, token_updater=None, **kwargs):
        """Construct a new OAuth 2 client session.

        :param client_id: Client id obtained during registration
        :param client: :class:`oauthlib.oauth2.Client` to be used. Default is
                       WebApplicationClient which is useful for any
                       hosted application but not mobile or desktop.
        :param scope: List of scopes you wish to request access to
        :param redirect_uri: Redirect URI you registered as callback
        :param token: Token dictionary, must include access_token
                      and token_type.
        :param state: State string used to prevent CSRF. This will be given
                      when creating the authorization url and must be supplied
                      when parsing the authorization response.
                      Can be either a string or a no argument callable.
        :auto_refresh_url: Refresh token endpoint URL, must be HTTPS. Supply
                           this if you wish the client to automatically refresh
                           your access tokens.
        :auto_refresh_kwargs: Extra arguments to pass to the refresh token
                              endpoint.
        :token_updater: Method with one argument, token, to be used to update
                        your token database on automatic token refresh. If not
                        set a TokenUpdated warning will be raised when a token
                        has been refreshed. This warning will carry the token
                        in its token argument.
        :param kwargs: Arguments to pass to the Session constructor.
        """
        super().__init__(**kwargs)
        self._client = client or WebApplicationClient(client_id, token=token)
        self.token = token or {}
        self.scope = scope
        self.redirect_uri = redirect_uri
        self.state = state or generate_token
        self._state = state
        self.auto_refresh_url = auto_refresh_url
        self.auto_refresh_kwargs = auto_refresh_kwargs or {}
        self.token_updater = token_updater

        # Allow customizations for non compliant providers through various
        # hooks to adjust requests and responses.
        self.compliance_hook = {
            'access_token_response': set(),
            'refresh_token_response': set(),
            'protected_request': set(),
        }

    def new_state(self):
        """Generates a state string to be used in authorizations."""
        try:
            self._state = self.state()
            _LOGGER.debug('Generated new state %s.', self._state)
        except TypeError:
            self._state = self.state
            _LOGGER.debug(
                'Re-using previously supplied state %s.', self._state)
        return self._state

    @property
    def client_id(self):
        """Get the client_id."""
        return getattr(self._client, "client_id", None)

    @client_id.setter
    def client_id(self, value):
        """Set the client_id."""
        self._client.client_id = value

    @client_id.deleter
    def client_id(self):
        """Remove the client_id."""
        del self._client.client_id

    @property
    def token(self):
        """Get the token."""
        return getattr(self._client, "token", None)

    @token.setter
    def token(self, value):
        """Set the token."""
        self._client.token = value
        self._client.populate_token_attributes(value)

    @property
    def access_token(self):
        """Get the access_token."""
        return getattr(self._client, "access_token", None)

    @access_token.setter
    def access_token(self, value):
        """Set the access_token."""
        self._client.access_token = value

    @access_token.deleter
    def access_token(self):
        """Remove the access_token."""
        del self._client.access_token

    @property
    def authorized(self):
        """Boolean that indicates whether this session has an OAuth token
        or not. If `self.authorized` is True, you can reasonably expect
        OAuth-protected requests to the resource to succeed. If
        `self.authorized` is False, you need the user to go through the OAuth
        authentication dance before OAuth-protected requests to the resource
        will succeed.
        """
        return bool(self.access_token)

    def authorization_url(self, url, state=None, **kwargs):
        """Form an authorization URL.

        :param url: Authorization endpoint url, must be HTTPS.
        :param state: An optional state string for CSRF protection. If not
                      given it will be generated for you.
        :param kwargs: Extra parameters to include.
        :return: authorization_url, state
        """
        state = state or self.new_state()
        return self._client.prepare_request_uri(
            url, redirect_uri=self.redirect_uri, scope=self.scope,
            state=state, **kwargs), state

    async def fetch_token(
            self, token_url, code=None, authorization_response=None,
            body='', auth=None, username=None, password=None, method='POST',
            timeout=None, headers=None, verify_ssl=True,
            post_payload_modifier=lambda x: x,  # proxies=None,
            **kwargs):
        """Generic method for fetching an access token from the token endpoint.

        If you are using the MobileApplicationClient you will want to use
        token_from_fragment instead of fetch_token.

        :param token_url: Token endpoint URL, must use HTTPS.
        :param code: Authorization code (used by WebApplicationClients).
        :param authorization_response: Authorization response URL, the callback
                                       URL of the request back to you. Used by
                                       WebApplicationClients instead of code.
        :param body: Optional application/x-www-form-urlencoded body to add the
                     include in the token request. Prefer kwargs over body.
        :param auth: An auth tuple or method as accepted by requests.
        :param username: Username used by LegacyApplicationClients.
        :param password: Password used by LegacyApplicationClients.
        :param method: The HTTP method used to make the request. Defaults
                       to POST, but may also be GET. Other methods should
                       be added as needed.
        :param headers: Dict to default request headers with.
        :param timeout: Timeout of the request in seconds.
        :param verify_ssl: Verify SSL certificate.
        :param post_payload_modifier: Allow to modify the payload to post (e.g. for JSON)
        :param kwargs: Extra parameters to include in the token request.
        :return: A token dict
        """
        if not is_secure_transport(token_url):
            raise InsecureTransportError()

        if not code and authorization_response:
            print('-- response', authorization_response)
            self._client.parse_request_uri_response(
                str(authorization_response), state=self._state)
            code = self._client.code
            print('--code', code)
        elif not code and isinstance(self._client, WebApplicationClient):
            code = self._client.code
            if not code:
                raise ValueError('Please supply either code or '
                                 'authorization_response parameters.')

        body = self._client.prepare_request_body(
            code=code, body=body, redirect_uri=self.redirect_uri,
            username=username, password=password, **kwargs)

        client_id = kwargs.get('client_id', '')
        if auth is None:
            if client_id:
                _LOGGER.debug(
                    'Encoding client_id "%s" with client_secret as '
                    'Basic auth credentials.', client_id)
                client_secret = kwargs.get('client_secret', '')
                client_secret = client_secret if client_secret is not None \
                    else ''
                auth = aiohttp.BasicAuth(
                    login=client_id, password=client_secret)
            elif username:
                if password is None:
                    raise ValueError('Username was supplied, but no password.')
                _LOGGER.debug(
                    'Encoding username, password as Basic auth credentials.')
                auth = aiohttp.BasicAuth(login=username, password=password)

        headers = headers or {
            'Accept': 'application/json',
            'Content-Type': 'application/x-www-form-urlencoded;charset=UTF-8',
        }
        self.token = {}

        async def _get_token(resp, source):
            _LOGGER.debug(
                'Request to fetch token completed with status %s.',
                resp.status)
            _LOGGER.debug('Request headers were %s', headers)
            _LOGGER.debug('Request body was %s', body)
            text = await resp.text()
            _LOGGER.debug(
                'Response headers were %s and content %s.',
                resp.headers, text)
            self._invoke_hooks(resp, 'access_token_response')

            self._client.parse_request_body_response(text, scope=self.scope)
            self.token = self._client.token
            _LOGGER.debug('Obtained token %s. From %s [%s]', self.token,
                          source, body)
            return self.token

        if method.upper() == 'POST':
            async with self.post(
                    token_url, data=post_payload_modifier(dict(urldecode(body))),
                    timeout=timeout, headers=headers, auth=auth,
                    verify_ssl=verify_ssl) as resp:  # proxies=proxies
                return await _get_token(resp, 'body')

        if method.upper() == 'GET':
            # if method is not 'POST', switch body to querystring and GET
            async with self.get(
                    token_url, params=dict(urldecode(body)),
                    timeout=timeout, headers=headers, auth=auth,
                    verify_ssl=verify_ssl) as resp:  # proxies=proxies
                return await _get_token(resp, 'querystring')

        raise ValueError('The method kwarg must be POST or GET.')

    def token_from_fragment(self, authorization_response):
        """Parse token from the URI fragment, used by MobileApplicationClients.

        :param authorization_response: The full URL of the redirect back to you
        :return: A token dict
        """
        self._client.parse_request_uri_response(
            authorization_response, state=self._state)
        self.token = self._client.token
        return self.token

    async def refresh_token(
            self, token_url, refresh_token=None, body='', auth=None,
            timeout=None, headers=None, verify_ssl=True,
            post_payload_modifier=lambda x: x,  # proxies=None,
            **kwargs):
        """Fetch a new access token using a refresh token.

        :param token_url: The token endpoint, must be HTTPS.
        :param refresh_token: The refresh_token to use.
        :param body: Optional application/x-www-form-urlencoded body to add the
                     include in the token request. Prefer kwargs over body.
        :param auth: An auth tuple or method as accepted by requests.
        :param timeout: Timeout of the request in seconds.
        :param verify_ssl: Verify SSL certificate.
        :param post_payload_modifier: Allow to modify the payload to post (e.g. for JSON)
        :param kwargs: Extra parameters to include in the token request.
        :return: A token dict
        """
        if not token_url:
            raise ValueError('No token endpoint set for auto_refresh.')

        if not is_secure_transport(token_url):
            raise InsecureTransportError()

        refresh_token = refresh_token or self.token.get('refresh_token')

        _LOGGER.debug(
            'Adding auto refresh key word arguments %s.',
            self.auto_refresh_kwargs)
        kwargs.update(self.auto_refresh_kwargs)
        body = self._client.prepare_refresh_body(
            body=body, refresh_token=refresh_token, scope=self.scope, **kwargs)
        _LOGGER.debug('Prepared refresh token request body %s', body)

        if headers is None:
            headers = {
                'Accept': 'application/json',
                'Content-Type': (
                    'application/x-www-form-urlencoded;charset=UTF-8'
                ),
            }

        async with self.post(
                token_url, data=post_payload_modifier(dict(urldecode(body))), auth=auth,
                timeout=timeout, headers=headers, verify_ssl=verify_ssl,
                withhold_token=True) as resp:  # proxies=proxies

            _LOGGER.debug(
                'Request to refresh token completed with status %s.',
                resp.status)
            text = await resp.text()
            _LOGGER.debug(
                'Response headers were %s and content %s.',
                resp.headers, text)

            self._invoke_hooks(text, 'refresh_token_response')

            self.token = self._client.parse_request_body_response(
                text, scope=self.scope)
            if 'refresh_token' not in self.token:
                _LOGGER.debug('No new refresh token given. Re-using old.')
                self.token['refresh_token'] = refresh_token
            return self.token

    async def _request(
            self, method, url, *, data=None, headers=None,
            withhold_token=False, client_id=None, client_secret=None,
            **kwargs):  # pylint: disable=arguments-differ
        """Intercept all requests and add the OAuth 2 token if present."""
        if not is_secure_transport(url):
            raise InsecureTransportError()
        if self.token and not withhold_token:
            self._invoke_hooks((url, headers, data), 'protected_request')
            _LOGGER.debug('Adding token %s to request.', self.token)
            try:
                url, headers, data = self._client.add_token(
                    url, http_method=method, body=data, headers=headers)
            # Attempt to retrieve and save new access token if expired
            except TokenExpiredError as token_expired_error:
                if self.auto_refresh_url:
                    _LOGGER.debug(
                        'Auto refresh is set, attempting to refresh at %s.',
                        self.auto_refresh_url)

                    # We mustn't pass auth twice.
                    auth = kwargs.pop('auth', None)
                    if client_id and client_secret and (auth is None):
                        _LOGGER.debug(
                            "Encoding client_id '%s' with client_secret as "
                            "Basic auth credentials.", client_id)
                        auth = aiohttp.BasicAuth(
                            login=client_id, password=client_secret)
                    token = await self.refresh_token(
                        self.auto_refresh_url, auth=auth, **kwargs)
                    if self.token_updater:
                        _LOGGER.debug(
                            "Updating token to %s using %s.",
                            token, self.token_updater)
                        self.token_updater(token)
                        url, headers, data = self._client.add_token(
                            url, http_method=method, body=data,
                            headers=headers)
                    else:
                        raise TokenUpdated(token) from token_expired_error
                else:
                    raise

        _LOGGER.debug('Requesting url %s using method %s.', url, method)
        _LOGGER.debug('Supplying headers %s and data %s', headers, data)
        _LOGGER.debug('Passing through key word arguments %s.', kwargs)
        return await super()._request(
            method, url, headers=headers, data=data, **kwargs)

    def _invoke_hooks(self, reqres, hook_type):
        _LOGGER.debug(
            "Invoking %d %s hooks.", len(self.compliance_hook[hook_type]),
            hook_type)
        for hook in self.compliance_hook[hook_type]:
            _LOGGER.debug("Invoking hook %s.", hook)
            reqres = hook(reqres)
        return reqres

    def register_compliance_hook(self, hook_type, hook):
        """Register a hook for request/response tweaking.

        Available hooks are:
            access_token_response invoked before token parsing.
            refresh_token_response invoked before refresh token parsing.
            protected_request invoked before making a request.

        If you find a new hook is needed please send a GitHub PR request
        or open an issue.
        """
        if hook_type not in self.compliance_hook:
            raise ValueError("Hook type {} is not in {}.".format(
                hook_type, self.compliance_hook))
        self.compliance_hook[hook_type].add(hook)



"""Python Module to use with SNOO Smart Sleeper Bassinet

Contains classes to authenticate against the Happiest Baby API and query various
entities and expose their functionality.
"""

from .auth_session import SnooAuthSession
from .snoo import Snoo
from .pubnub import SnooPubNub
from .models import (User,
                     Device,
                     Baby,
                     SSID,
                     Picture,
                     Settings,
                     ResponsivenessLevel,
                     MinimalLevelVolume,
                     SoothingLevelVolume,
                     MinimalLevel,
                     Sex,
                     LastSession,
                     SessionLevel,
                     AggregatedSession,
                     AggregatedSessionItem,
                     SessionItemType,
                     AggregatedSessionAvg,
                     ActivityState)

__all__ = ['SnooAuthSession',
           'Snoo',
           'SnooPubNub',
           'User',
           'Device',
           'Baby',
           'SSID',
           'Picture',
           'Settings',
           'ResponsivenessLevel',
           'MinimalLevelVolume',
           'SoothingLevelVolume',
           'MinimalLevel',
           'Sex',
           'LastSession',
           'SessionLevel',
           'AggregatedSession',
           'AggregatedSessionItem',
           'SessionItemType',
           'AggregatedSessionAvg',
           'ActivityState']



"""PySnoo PubNub Interface."""
import asyncio
import logging
from typing import Callable, Optional, List

from pubnub.callbacks import SubscribeCallback
from pubnub.pnconfiguration import PNConfiguration
from pubnub.pubnub_asyncio import PubNubAsyncio, utils

from .models import ActivityState, SessionLevel
from .const import SNOO_PUBNUB_PUBLISH_KEY, SNOO_PUBNUB_SUBSCRIBE_KEY

_LOGGER = logging.getLogger(__name__)


class SnooSubscribeListener(SubscribeCallback):
    """Snoo Subscription Listener Class"""

    def __init__(self, callback: Callable[[ActivityState], None]):
        """Initialize the Snoo Subscription Listener"""
        self.connected_event = asyncio.Event()
        self.disconnected_event = asyncio.Event()
        self._callback = callback

    def status(self, pubnub, status):
        """PubNub Status Callback Implementation"""
        if utils.is_subscribed_event(status) and not self.connected_event.is_set():
            self.connected_event.set()
            self.disconnected_event.clear()
        elif utils.is_unsubscribed_event(status) and not self.disconnected_event.is_set():
            self.disconnected_event.set()
            self.connected_event.clear()
        elif status.is_error():
            _LOGGER.error('Error in Snoo PubNub Listener of Category: %s', status.category)

    def message(self, pubnub, message):
        """PubNub Message Callback Implementation"""
        self._callback(ActivityState.from_dict(message.message))

    def presence(self, pubnub, presence):
        """PubNub Presence Callback Implementation"""

    def is_connected(self):
        """Returns true if the listener is currently connected to an active subscription"""
        return self.connected_event.is_set()

    async def wait_for_connect(self):
        """Async utility function that waits for subscription connect."""
        if not self.connected_event.is_set():
            await self.connected_event.wait()

    async def wait_for_disconnect(self):
        """Async utility function that waits for subscription disconnect."""
        if not self.disconnected_event.is_set():
            await self.disconnected_event.wait()


class SnooPubNub:
    """A Python Abstraction for Snoos PubNub Interface."""
    # pylint: disable=too-few-public-methods,fixme

    def __init__(self,
                 access_token: str,
                 serial_number: str,
                 uuid: str,
                 custom_event_loop=None):
        """Initialize the Snoo PubNub object."""
        self.config = self._setup_pnconfig(access_token, uuid)
        self.serial_number = serial_number
        self._activiy_channel = 'ActivityState.{}'.format(serial_number)
        self._controlcommand_channel = 'ControlCommand.{}'.format(serial_number)
        self._pubnub = PubNubAsyncio(self.config, custom_event_loop=custom_event_loop)
        self._listener = SnooSubscribeListener(self._activy_state_callback)
        # Add listener
        self._pubnub.add_listener(self._listener)
        self._external_listeners: List[Callable[[ActivityState], None]] = []

    @staticmethod
    def _setup_pnconfig(access_token, uuid):
        """Generate Setup"""
        pnconfig = PNConfiguration()
        pnconfig.subscribe_key = SNOO_PUBNUB_SUBSCRIBE_KEY
        pnconfig.publish_key = SNOO_PUBNUB_PUBLISH_KEY
        pnconfig.uuid = uuid
        pnconfig.auth_key = access_token
        pnconfig.ssl = True
        return pnconfig

    def add_listener(self, update_callback: Callable[[ActivityState], None]) -> Callable[[], None]:
        """Add a AcitivyState Listener to the SnooPubNub Entity and returns a remove_listener CB for that listener"""
        self._external_listeners.append(update_callback)

        def remove_listener_cb() -> None:
            """Remove listener."""
            self.remove_listener(update_callback)

        return remove_listener_cb

    def remove_listener(self, update_callback: Callable[[ActivityState], None]) -> None:
        """Remove data update."""
        self._external_listeners.remove(update_callback)

    def _activy_state_callback(self, state: ActivityState):
        """Internal Callback of SnooSubscribeListener"""
        for update_callback in self._external_listeners:
            update_callback(state)

    def subscribe(self):
        """Subscribe to Snoo Activity Channel"""
        if self._listener.is_connected():
            _LOGGER.warning('Trying to subscribe PubNub instance that is already subscribed to %s',
                            self._activiy_channel)
            return

        self._pubnub.subscribe().channels([
            self._activiy_channel
        ]).execute()

    async def subscribe_and_await_connect(self):
        """Subscribe to Snoo Activity Channel and await connect"""
        self.subscribe()
        await self._listener.wait_for_connect()

    def unsubscribe(self):
        """Unsubscribe to Snoo Activity Channel"""
        if not self._listener.is_connected():
            _LOGGER.warning('Trying to unsubscribe PubNub instance that is NOT subscribed to %s', self._activiy_channel)
            return

        self._pubnub.unsubscribe().channels(
            self._activiy_channel
        ).execute()

    async def unsubscribe_and_await_disconnect(self):
        """Unsubscribe to Snoo Activity Channel and await disconnect"""
        self.unsubscribe()
        await self._listener.wait_for_disconnect()

    async def history(self, count=1):
        """Retrieve number of count historic messages"""
        envelope = await self._pubnub.history().channel(
            self._activiy_channel
        ).count(count).future()
        return [ActivityState.from_dict(item.entry) for item in envelope.result.messages]

    async def publish(self, message):
        """Publish a message to the Snoo control command channel"""
        task = await self._pubnub.publish().channel(
            self._controlcommand_channel).message(message).future()
        return task

    async def publish_goto_state(self, level: SessionLevel, hold: Optional[bool] = None):
        """Publish a message a go_to_state command to the Snoo control command channel"""
        msg = {
            'command': 'go_to_state',
            'state': level.value
        }
        if hold is not None:
            msg['hold'] = 'on' if hold else 'off'
        return await self.publish(msg)

    async def publish_start(self):
        """Publish a message a start_snoo command to the Snoo control command channel"""
        return await self.publish({
            'command': 'start_snoo'
        })

    async def stop(self):
        """Stop and Cleanup the Async Pubnub Utility"""
        # pylint: disable=protected-access
        # Workaround until PR is accepted:
        # https://github.com/pubnub/python/pull/99
        # self._pubnub.stop()
        await self._pubnub._session.close()
        if self._pubnub._subscription_manager is not None:
            self._pubnub._subscription_manager.stop()



"""PySnoo Constants."""

# Base Headers
BASE_HEADERS = {
    'User-Agent': 'okhttp/4.7.2',
}

# Snoo API endpoints
SNOO_API_URI = 'https://snoo-api.happiestbaby.com'

# OAuth-related
OAUTH_LOGIN_ENDPOINT = SNOO_API_URI + '/us/login/'
OAUTH_TOKEN_REFRESH_ENDPOINT = SNOO_API_URI + '/us/refresh/'
OAUTH_CLIENT_ID = 'snoo_client'
OAUTH_SCOPE = ['offline_access']

SNOO_REGISTER_ENDPOINT = SNOO_API_URI + '/us/register/'
SNOO_ME_ENDPOINT = SNOO_API_URI + '/us/me/'
SNOO_DEVICES_ENDPOINT = SNOO_API_URI + '/ds/me/devices/'
SNOO_BABY_ENDPOINT = SNOO_API_URI + '/us/v3/me/baby/'
SNOO_SESSIONS_LAST_ENDPOINT = SNOO_API_URI + '/ss/v2/sessions/last/'
SNOO_SESSIONS_AGGREGATED_ENDPOINT = SNOO_API_URI + '/ss/v2/sessions/aggregated/'
SNOO_SESSIONS_AGGREGATED_AVG_ENDPOINT = SNOO_API_URI + '/ss/v2/babies/{}/sessions/aggregated/avg/'
SNOO_SESSIONS_TOTAL_TIME_ENDPOINT = SNOO_API_URI + '/ss/v2/babies/{}/sessions/total-time/'

# Snoo Pubnub Variables
SNOO_PUBNUB_SUBSCRIBE_KEY = "sub-c-97bade2a-483d-11e6-8b3b-02ee2ddab7fe"
SNOO_PUBNUB_PUBLISH_KEY = "pub-c-699074b0-7664-4be2-abf8-dcbb9b6cd2bf"

DATETIME_FMT_AGGREGATED_SESSION = '%Y-%m-%d %H:%M:%S.%f'



"""PySnoo Data Models."""
from typing import List, Optional
from dataclasses import dataclass
from datetime import datetime, date, timedelta, timezone
from enum import Enum

from .const import DATETIME_FMT_AGGREGATED_SESSION


# from: https://github.com/ctalkington/python-sonarr/blob/master/sonarr/models.py
def dt_str_to_dt(dt_str: str) -> datetime:
    """Convert ISO-8601 datetime string to datetime object."""
    if dt_str is None:
        return None

    utc = False

    if "Z" in dt_str:
        utc = True
        dt_str = dt_str[:-1]

    if "." in dt_str:
        # Python doesn't support long microsecond values
        ts_bits = dt_str.split(".", 1)
        dt_str = "{}.{}".format(ts_bits[0], ts_bits[1][:3])
        fmt = "%Y-%m-%dT%H:%M:%S.%f"
    else:
        fmt = "%Y-%m-%dT%H:%M:%S"

    if utc:
        dt_str += "Z"
        fmt += "%z"

    return datetime.strptime(dt_str, fmt)


def dt_to_dt_str(dt_value: datetime) -> str:
    """Convert datetime object to ISO-8601 datetime string object."""
    if dt_value is None:
        return None

    return dt_value.isoformat(timespec='milliseconds').replace('+00:00', 'Z')


@dataclass(frozen=True)
class User:
    """Object holding the user information from Snoo."""

    email: str
    given_name: str
    region: str
    surname: str
    user_id: str

    @staticmethod
    def from_dict(data: dict):
        """Return user object from dict."""
        return User(
            email=data.get("email", ""),
            given_name=data.get("givenName", ""),
            region=data.get("region", ""),
            surname=data.get("surname", ""),
            user_id=data.get("userId", ""),
        )

    def to_dict(self):
        """Return dict from Object"""
        return {
            "email": self.email,
            "givenName": self.given_name,
            "region": self.region,
            "surname": self.surname,
            "userId": self.user_id
        }


@dataclass(frozen=True)
class SSID:
    """Object holding ssid information."""

    name: str
    updated_at: datetime

    @staticmethod
    def from_dict(data: dict):
        """Return SSID object from dict."""
        return SSID(
            name=data.get("name", ""),
            updated_at=dt_str_to_dt(data.get("updatedAt", None))
        )

    def to_dict(self):
        """Return dict from Object"""
        return {
            "name": self.name,
            "updatedAt": dt_to_dt_str(self.updated_at)
        }


@dataclass(frozen=True)
class Device:
    """Object holding Snoo device information."""
    baby: str  # ID of baby
    created_at: datetime
    firmware_update_date: datetime
    firmware_version: str
    last_provision_success: datetime
    last_ssid: SSID
    serial_number: str
    timezone: str
    updated_at: datetime

    @staticmethod
    def from_dict(data: dict):
        """Return device object from dict."""
        return Device(
            baby=data.get("baby", ""),
            created_at=dt_str_to_dt(data.get("createdAt", None)),
            firmware_update_date=dt_str_to_dt(data.get("firmwareUpdateDate", None)),
            firmware_version=data.get("firmwareVersion", ""),
            last_provision_success=dt_str_to_dt(data.get("lastProvisionSuccess", None)),
            last_ssid=SSID.from_dict(data.get("lastSSID", {})),
            serial_number=data.get("serialNumber", ""),
            timezone=data.get("timezone", ""),
            updated_at=dt_str_to_dt(data.get("updatedAt", None)),
        )

    def to_dict(self):
        """Return dict from Object"""
        return {
            "baby": self.baby,
            "createdAt": dt_to_dt_str(self.created_at),
            "firmwareUpdateDate": dt_to_dt_str(self.firmware_update_date),
            "firmwareVersion": self.firmware_version,
            "lastProvisionSuccess": dt_to_dt_str(self.last_provision_success),
            "lastSSID": self.last_ssid.to_dict(),
            "serialNumber": self.serial_number,
            "timezone": self.timezone,
            "updatedAt": dt_to_dt_str(self.updated_at)
        }


@dataclass(frozen=True)
class Picture:
    """Object holding picture information."""

    id: str  # pylint: disable=invalid-name
    mime: str
    encoded: bool
    updated_at: datetime

    @staticmethod
    def from_dict(data: dict):
        """Return picture object from dict."""
        return Picture(
            id=data.get("id", ""),
            mime=data.get("mime", ""),
            encoded=data.get("encoded", False),
            updated_at=dt_str_to_dt(data.get("updatedAt", None)),
        )

    def to_dict(self):
        """Return dict from Object"""
        return {
            "id": self.id,
            "mime": self.mime,
            "encoded": self.encoded,
            "updatedAt": dt_to_dt_str(self.updated_at)
        }


class ResponsivenessLevel(Enum):
    """Enum for ResponsivenessLevel"""
    VERY_LOW = 'lvl-2'
    LOW = 'lvl-1'
    NORMAL = 'lvl0'
    HIGH = 'lvl+1'
    VERY_HIGH = 'lvl+2'


class MinimalLevelVolume(Enum):
    """Enum for MinimalLevelVolume"""
    VERY_LOW = 'lvl-2'
    LOW = 'lvl-1'
    NORMAL = 'lvl0'
    HIGH = 'lvl+1'
    VERY_HIGH = 'lvl+2'


class SoothingLevelVolume(Enum):
    """Enum for SoothingLevelVolume"""
    NORMAL = 'lvl0'
    HIGH = 'lvl+1'
    VERY_HIGH = 'lvl+2'


class MinimalLevel(Enum):
    """Enum for MinimalLevel"""
    BASELINE = 'baseline'
    LEVEL1 = 'level1'
    LEVEL2 = 'level2'


class Sex(Enum):
    """Enum for Sex"""
    MALE = 'Male'
    FEMALE = 'Female'


@dataclass(frozen=True)
class Settings:
    """Object holding Snoo Settings information."""

    responsiveness_level: ResponsivenessLevel
    minimal_level_volume: MinimalLevelVolume
    soothing_level_volume: SoothingLevelVolume
    minimal_level: MinimalLevel
    motion_limiter: bool
    weaning: bool
    car_ride_mode: bool
    offline_lock: bool
    # App restriction is 5-12 (am)
    daytime_start: int

    @staticmethod
    def from_dict(data: dict):
        """Return device object from dict."""
        return Settings(
            responsiveness_level=ResponsivenessLevel(data.get("responsivenessLevel", "lvl0")),
            minimal_level_volume=MinimalLevelVolume(data.get("minimalLevelVolume", "lvl0")),
            soothing_level_volume=SoothingLevelVolume(data.get("soothingLevelVolume", "lvl0")),
            minimal_level=MinimalLevel(data.get("minimalLevel", "baseline")),
            motion_limiter=data.get("motionLimiter", False),
            weaning=data.get("weaning", False),
            car_ride_mode=data.get("carRideMode", False),
            offline_lock=data.get("offlineLock", False),
            daytime_start=data.get("daytimeStart", 7)
        )

    def to_dict(self):
        """Return dict from Object"""
        return {
            "responsivenessLevel": self.responsiveness_level.value,
            "minimalLevelVolume": self.minimal_level_volume.value,
            "soothingLevelVolume": self.soothing_level_volume.value,
            "minimalLevel": self.minimal_level.value,
            "motionLimiter": self.motion_limiter,
            "weaning": self.weaning,
            "carRideMode": self.car_ride_mode,
            "offlineLock": self.offline_lock,
            "daytimeStart": self.daytime_start
        }


@dataclass(frozen=True)
class Baby:
    """Object for Snoo Baby information."""

    baby: str  # ID of baby
    baby_name: str
    birth_date: date
    created_at: datetime
    disabled_limiter: bool
    pictures: List[Picture]
    # Can be None
    preemie: int
    settings: Settings
    # Can be None
    sex: Sex
    updated_at: datetime
    updated_by_user_at: datetime

    @staticmethod
    def from_dict(data: dict):
        """Return baby object from dict."""
        birth_date = dt_str_to_dt(data.get("birthDate", None))
        if birth_date is not None:
            birth_date = birth_date.date()

        return Baby(
            baby=data.get("_id", ""),
            baby_name=data.get("babyName", ""),
            birth_date=birth_date,
            created_at=dt_str_to_dt(data.get("createdAt", None)),
            disabled_limiter=data.get("disabledLimiter", False),
            pictures=[Picture.from_dict(p) for p in data.get("pictures", [])],
            settings=Settings.from_dict(data.get("settings", {})),
            preemie=data.get("preemie"),
            sex=data.get("sex"),
            updated_at=dt_str_to_dt(data.get("updatedAt", None)),
            updated_by_user_at=dt_str_to_dt(data.get("updatedByUserAt", None)),
        )

    def to_dict(self):
        """Return dict from Object"""
        birth_date = self.birth_date
        if birth_date is not None:
            birth_date = birth_date.isoformat()

        return {
            "baby": self.baby,
            "babyName": self.baby_name,
            "birthDate": birth_date,
            "createdAt": dt_to_dt_str(self.created_at),
            "disabledLimiter": self.disabled_limiter,
            "pictures": [item.to_dict() for item in self.pictures],
            "settings": self.settings.to_dict(),
            "preemie": self.preemie,
            "sex": self.sex,
            "updatedAt": dt_to_dt_str(self.updated_at),
            "updatedByUserAt": dt_to_dt_str(self.updated_by_user_at),
        }


class SessionLevel(Enum):
    """Enum for SessionLevel"""
    ONLINE = 'ONLINE'
    BASELINE = 'BASELINE'
    WEANING_BASELINE = 'WEANING_BASELINE'
    LEVEL1 = 'LEVEL1'
    LEVEL2 = 'LEVEL2'
    LEVEL3 = 'LEVEL3'
    LEVEL4 = 'LEVEL4'
    NONE = 'NONE'
    PRETIMEOUT = 'PRETIMEOUT'

    def is_active_level(self):
        """Returns true if the Enum value represents an active level."""
        return self in [SessionLevel.BASELINE,
                        SessionLevel.WEANING_BASELINE,
                        SessionLevel.LEVEL1,
                        SessionLevel.LEVEL2,
                        SessionLevel.LEVEL3,
                        SessionLevel.LEVEL4]


class SessionItemType(Enum):
    """Enum for SessionItemType"""
    ASLEEP = 'asleep'
    SOOTHING = 'soothing'
    AWAKE = 'awake'


@dataclass(frozen=True)
class LastSession:
    """Object for Snoo LastSession information."""

    end_time: datetime
    levels: List[SessionLevel]
    start_time: datetime

    @property
    def current_status_duration(self) -> timedelta:
        """Return the duration spent in current status"""
        if self.end_time:
            return datetime.now(timezone.utc) - self.end_time

        return datetime.now(timezone.utc) - self.start_time

    @property
    def current_status(self) -> SessionItemType:
        """Return the current status"""
        if self.end_time:
            return SessionItemType.AWAKE

        if self.levels[-1] in [SessionLevel.BASELINE, SessionLevel.WEANING_BASELINE]:
            return SessionItemType.ASLEEP

        return SessionItemType.SOOTHING

    @staticmethod
    def from_dict(data: dict):
        """Return LastSession object from dict."""
        return LastSession(
            end_time=dt_str_to_dt(data.get("endTime", None)),
            levels=[SessionLevel(item['level']) for item in data.get("levels", [])],
            start_time=dt_str_to_dt(data.get("startTime", None)),
        )

    def to_dict(self):
        """Return dict from Object"""

        return {
            "endTime": dt_to_dt_str(self.end_time),
            "levels": [item.value for item in self.levels],
            "startTime": dt_to_dt_str(self.start_time),
            "currentStatus": self.current_status.value,
            "currentStatusDuration": str(self.current_status_duration)
        }


@dataclass(frozen=True)
class AggregatedSessionItem:
    """Object for Snoo AggregatedSessionItem information."""

    is_active: bool
    session_id: str
    start_time: datetime
    state_duration: timedelta
    type: SessionItemType

    @staticmethod
    def from_dict(data: dict):
        """Return AggregatedSessionItem object from dict."""

        start_time = data.get("startTime")
        if start_time is not None:
            start_time = datetime.strptime(start_time, DATETIME_FMT_AGGREGATED_SESSION)

        return AggregatedSessionItem(
            is_active=data.get("isActive", False),
            session_id=data.get("sessionId", ""),
            start_time=start_time,
            state_duration=timedelta(seconds=data.get("stateDuration", 0)),
            type=SessionItemType(data.get("type"))
        )

    def to_dict(self):
        """Return dict from Object"""
        return {
            "isActive": self.is_active,
            "sessionId": self.session_id,
            "startTime": dt_to_dt_str(self.start_time),
            "stateDuration": str(self.state_duration),
            "type": self.type.value
        }


@dataclass(frozen=True)
class AggregatedSession:
    """Object for Snoo AggregatedSession information."""

    day_sleep: timedelta
    levels: List[AggregatedSessionItem]
    longest_sleep: timedelta
    naps: int
    night_sleep: timedelta
    night_wakings: int
    timezone: str
    total_sleep: timedelta

    @staticmethod
    def from_dict(data: dict):
        """Return AggregatedSession object from dict."""
        return AggregatedSession(
            day_sleep=timedelta(seconds=data.get("daySleep", 0)),
            levels=[AggregatedSessionItem.from_dict(item) for item in data.get("levels", [])],
            longest_sleep=timedelta(seconds=data.get("longestSleep", 0)),
            naps=data.get("naps", 0),
            night_sleep=timedelta(seconds=data.get("nightSleep", 0)),
            night_wakings=data.get("nightWakings", 0),
            timezone=data.get("timezone"),
            total_sleep=timedelta(seconds=data.get("totalSleep", 0)),
        )

    def to_dict(self):
        """Return dict from Object"""
        return {
            "daySleep": str(self.day_sleep),
            "levels": [item.to_dict() for item in self.levels],
            "longestSleep": str(self.longest_sleep),
            "naps": self.naps,
            "nightSleep": str(self.night_sleep),
            "nightWakings": self.night_wakings,
            "timezone": self.timezone,
            "totalSleep": str(self.total_sleep),
        }


class AggregatedSessionInterval(Enum):
    """Enum for AggregatedSessionInterval"""
    WEEK = 'week'
    MONTH = 'month'


@dataclass(frozen=True)
class AggregatedDays:
    """Object for Snoo AggregatedDays information."""
    total_sleep: List[timedelta]
    day_sleep: List[timedelta]
    night_sleep: List[timedelta]
    longest_sleep: List[timedelta]
    night_wakings: List[int]

    @staticmethod
    def from_dict(data: dict):
        """Return AggregatedDays object from dict."""
        if data is None:
            return None

        return AggregatedDays(
            total_sleep=[timedelta(seconds=item) for item in data.get("totalSleep", [])],
            day_sleep=[timedelta(seconds=item) for item in data.get("daySleep", [])],
            night_sleep=[timedelta(seconds=item) for item in data.get("nightSleep", [])],
            longest_sleep=[timedelta(seconds=item) for item in data.get("longestSleep", [])],
            night_wakings=data.get("nightWakings", [])
        )

    def to_dict(self):
        """Return dict from Object"""
        return {
            "totalSleep": [str(item) for item in self.total_sleep],
            "daySleep": [str(item) for item in self.day_sleep],
            "nightSleep": [str(item) for item in self.night_sleep],
            "longestSleep": [str(item) for item in self.longest_sleep],
            "nightWakings": self.night_wakings,
        }


@dataclass(frozen=True)
class AggregatedSessionAvg:
    """Object for Snoo AggregatedSessionAvg information."""

    total_sleep_avg: timedelta
    day_sleep_avg: timedelta
    night_sleep_avg: timedelta
    longest_sleep_avg: timedelta
    night_wakings_avg: float
    days: Optional[AggregatedDays]

    @staticmethod
    def from_dict(data: dict):
        """Return AggregatedSessionAvg object from dict."""
        return AggregatedSessionAvg(
            total_sleep_avg=timedelta(seconds=data.get("totalSleepAVG", 0)),
            day_sleep_avg=timedelta(seconds=data.get("daySleepAVG", 0)),
            night_sleep_avg=timedelta(seconds=data.get("nightSleepAVG", 0)),
            longest_sleep_avg=timedelta(seconds=data.get("longestSleepAVG", 0)),
            night_wakings_avg=data.get("nightWakingsAVG", 0.0),
            days=AggregatedDays.from_dict(data.get("days")),
        )

    def to_dict(self):
        """Return dict from Object"""
        days = self.days
        if days is not None:
            days = days.to_dict()

        return {
            "totalSleepAVG": str(self.total_sleep_avg),
            "daySleepAVG": str(self.day_sleep_avg),
            "nightSleepAVG": str(self.night_sleep_avg),
            "longestSleepAVG": str(self.longest_sleep_avg),
            "nightWakingsAVG": self.night_wakings_avg,
            "days": days
        }


@dataclass(frozen=True)
class Signal:
    """Object for Snoo Signal information."""
    rssi: int
    strength: int

    @staticmethod
    def from_dict(data: dict):
        """Return AggregatedSessionAvg object from dict."""
        return Signal(**data)

    def to_dict(self):
        """Return dict from Object"""
        return vars(self)


@dataclass(frozen=True)
class StateMachine:
    """Object for Snoo StateMachine information."""

    up_transition: SessionLevel
    since_session_start: timedelta
    sticky_white_noise: bool
    weaning: bool
    time_left: timedelta
    session_id: str
    state: SessionLevel
    is_active_session: bool
    down_transition: SessionLevel
    hold: bool
    audio: bool

    @staticmethod
    def from_dict(data: dict):
        """Return StateMachine object from dict."""
        time_left = data.get("time_left")
        if time_left and time_left >= 0:
            time_left = timedelta(seconds=time_left)
        else:
            time_left = None

        since_session_start = data.get("since_session_start_ms")
        if since_session_start and since_session_start >= 0:
            since_session_start = timedelta(milliseconds=since_session_start)
        else:
            since_session_start = None

        return StateMachine(
            up_transition=SessionLevel(data.get("up_transition", SessionLevel.NONE.value)),
            since_session_start=since_session_start,
            sticky_white_noise=data.get("sticky_white_noise") == 'on',
            weaning=data.get("weaning") == 'on',
            time_left=time_left,
            session_id=data.get("session_id"),
            state=SessionLevel(data.get("state", SessionLevel.NONE.value)),
            is_active_session=data.get("is_active_session") == 'true',
            down_transition=SessionLevel(data.get("down_transition", SessionLevel.NONE.value)),
            hold=data.get("hold") == 'on',
            audio=data.get("audio") == 'on'
        )

    def to_dict(self):
        """Return dict from Object"""
        return {
            'up_transition': self.up_transition.value,
            'since_session_start': None if self.since_session_start is None else str(self.since_session_start),
            'sticky_white_noise': self.sticky_white_noise,
            'weaning': self.weaning,
            'time_left': None if self.time_left is None else str(self.time_left),
            'session_id': self.session_id,
            'state': self.state.value,
            'is_active_session': self.is_active_session,
            'down_transition': self.down_transition.value,
            'hold': self.hold,
            'audio': self.audio
        }


class EventType(Enum):
    """Enum for EventType"""
    ACTIVITY = 'activity'
    CRY = 'cry'
    TIMER = 'timer'
    COMMAND = 'command'
    SAFETY_CLIP = 'safety_clip'


@dataclass(frozen=True)
class ActivityState:
    """Return AggregatedSessionAvg object from dict."""

    left_safety_clip: bool
    rx_signal: Signal
    right_safety_clip: bool
    sw_version: str
    event_time: datetime
    state_machine: StateMachine
    system_state: str
    event: EventType

    @staticmethod
    def from_dict(data: dict):
        """Return AggregatedSessionAvg object from dict."""
        return ActivityState(
            left_safety_clip=bool(data.get("left_safety_clip")),
            rx_signal=Signal.from_dict(data.get("rx_signal", {})),
            right_safety_clip=bool(data.get("right_safety_clip")),
            sw_version=data.get("sw_version"),
            event_time=datetime.utcfromtimestamp(data.get("event_time_ms") / 1000).replace(tzinfo=timezone.utc),
            state_machine=StateMachine.from_dict(data.get("state_machine", {})),
            system_state=data.get("system_state"),
            event=EventType(data.get("event", EventType.ACTIVITY.value)),
        )

    def to_dict(self):
        """Return dict from Object"""
        return {
            "left_safety_clip": self.left_safety_clip,
            "rx_signal": self.rx_signal.to_dict(),
            "right_safety_clip": self.right_safety_clip,
            "sw_version": self.sw_version,
            "event_time": dt_to_dt_str(self.event_time),
            "state_machine": self.state_machine.to_dict(),
            "system_state": self.system_state,
            "event": self.event.value,
        }



"""The main API class"""
from typing import List, Optional
from datetime import date, datetime, timedelta

from .const import (SNOO_ME_ENDPOINT,
                    SNOO_DEVICES_ENDPOINT,
                    SNOO_BABY_ENDPOINT,
                    SNOO_SESSIONS_LAST_ENDPOINT,
                    SNOO_SESSIONS_AGGREGATED_ENDPOINT,
                    SNOO_SESSIONS_AGGREGATED_AVG_ENDPOINT,
                    SNOO_SESSIONS_TOTAL_TIME_ENDPOINT,
                    DATETIME_FMT_AGGREGATED_SESSION)
from .auth_session import SnooAuthSession
from .models import (User, Device, Baby, Sex,
                     MinimalLevel,
                     MinimalLevelVolume,
                     ResponsivenessLevel,
                     SoothingLevelVolume,
                     LastSession,
                     AggregatedSession,
                     AggregatedSessionAvg,
                     AggregatedSessionInterval)


class Snoo:
    """A Python Abstraction object to Snoo Smart Sleeper Bassinett."""
    def __init__(self, auth: SnooAuthSession):
        """Initialize the Snoo object."""
        self.auth = auth

    async def get_me(self) -> User:
        """Return Information about the current User"""
        async with self.auth.get(SNOO_ME_ENDPOINT) as resp:
            assert resp.status == 200
            return User.from_dict(await resp.json())

    async def get_devices(self) -> List[Device]:
        """Return Information about the configured devices"""
        async with self.auth.get(SNOO_DEVICES_ENDPOINT) as resp:
            assert resp.status == 200
            resp_json = await resp.json()
            return [Device.from_dict(d) for d in resp_json]

    async def get_baby(self) -> Baby:
        """Return Information about the current User"""
        async with self.auth.get(SNOO_BABY_ENDPOINT) as resp:
            assert resp.status == 200
            return Baby.from_dict(await resp.json())

    async def get_last_session(self) -> LastSession:
        """Return Information about the last session"""
        async with self.auth.get(SNOO_SESSIONS_LAST_ENDPOINT) as resp:
            assert resp.status == 200
            return LastSession.from_dict(await resp.json())

    async def get_aggregated_session(self, start_time: datetime) -> AggregatedSession:
        """Return Information about the aggregated session

        This function returns information about the next 24h segment beginning from start_time.
        Note, start_time does not contain or respect a timezone property, but it will assume the
        timezone that is configured server-side.
        """
        url_params = {
            'startTime': start_time.strftime(DATETIME_FMT_AGGREGATED_SESSION)[:-3]
        }
        async with self.auth.get(SNOO_SESSIONS_AGGREGATED_ENDPOINT,
                                 params=url_params) as resp:
            assert resp.status == 200
            return AggregatedSession.from_dict(await resp.json())

    async def get_aggregated_session_avg(self,
                                         baby: str,
                                         start_time: datetime,
                                         interval: AggregatedSessionInterval = AggregatedSessionInterval.WEEK,
                                         days: bool = True) -> AggregatedSessionAvg:
        """Return Information about the aggregated session averages

        :param baby: ID of baby to get average for
        :param start_time: start_time of the interval (time is ignored)
        :param interval: week/month calculate average for a week or month interval
        :param days: true/false Include value for each day in response payload
        :return:
        """
        url_params = {
            'startTime': start_time.strftime(DATETIME_FMT_AGGREGATED_SESSION)[:-3],
            'interval': interval.value,
            'days': str(days).lower(),
        }
        async with self.auth.get(SNOO_SESSIONS_AGGREGATED_AVG_ENDPOINT.format(baby),
                                 params=url_params) as resp:
            assert resp.status == 200
            return AggregatedSessionAvg.from_dict(await resp.json())

    async def get_session_total_time(self,
                                     baby: str) -> timedelta:
        """Return Information about the total usage of a Snoo

        :param baby: ID of baby to get the total time for
        :return:
        """
        async with self.auth.get(SNOO_SESSIONS_TOTAL_TIME_ENDPOINT.format(baby)) as resp:
            assert resp.status == 200
            resp_json = await resp.json()
            return timedelta(seconds=resp_json.get('totalTime', 0))

    async def set_baby_info(self,
                            baby_name: str,
                            birth_date: date,
                            preemie: Optional[int],
                            sex: Optional[Sex]) -> Baby:
        """Updates and returns baby-related information"""

        if sex is not None:
            sex = sex.value
        request_payload = {
            'babyName': baby_name,
            'birthDate': birth_date.isoformat(),
            'preemie': preemie,
            'sex': sex
        }

        async with self.auth.patch(SNOO_BABY_ENDPOINT, json=request_payload) as resp:
            assert resp.status == 200
            return Baby.from_dict(await resp.json())

    async def set_minimal_level(self,
                                minimal_level: MinimalLevel) -> Baby:
        """Updates minimal_level setting and returns baby-related information"""
        request_payload = {
            'settings': {
                'minimalLevel': minimal_level.value
            }
        }

        async with self.auth.patch(SNOO_BABY_ENDPOINT, json=request_payload) as resp:
            assert resp.status == 200
            return Baby.from_dict(await resp.json())

    async def set_minimal_level_volume(self,
                                       minimal_level_volume: MinimalLevelVolume) -> Baby:
        """Updates minimal_level_volume setting and returns baby-related information"""
        request_payload = {
            'settings': {
                'minimalLevelVolume': minimal_level_volume.value
            }
        }

        async with self.auth.patch(SNOO_BABY_ENDPOINT, json=request_payload) as resp:
            assert resp.status == 200
            return Baby.from_dict(await resp.json())

    async def set_responsiveness_level(self,
                                       responsiveness_level: ResponsivenessLevel) -> Baby:
        """Updates responsiveness_level setting and returns baby-related information"""
        request_payload = {
            'settings': {
                'responsivenessLevel': responsiveness_level.value
            }
        }

        async with self.auth.patch(SNOO_BABY_ENDPOINT, json=request_payload) as resp:
            assert resp.status == 200
            return Baby.from_dict(await resp.json())

    async def set_soothing_level_volume(self,
                                        soothing_level_volume: SoothingLevelVolume) -> Baby:
        """Updates soothing_level_volume setting and returns baby-related information"""
        request_payload = {
            'settings': {
                'soothingLevelVolume': soothing_level_volume.value
            }
        }

        async with self.auth.patch(SNOO_BABY_ENDPOINT, json=request_payload) as resp:
            assert resp.status == 200
            return Baby.from_dict(await resp.json())

    async def set_motion_limiter(self,
                                 motion_limiter: bool) -> Baby:
        """Updates motion_limiter setting and returns baby-related information"""
        request_payload = {
            'settings': {
                'motionLimiter': motion_limiter
            }
        }

        async with self.auth.patch(SNOO_BABY_ENDPOINT, json=request_payload) as resp:
            assert resp.status == 200
            return Baby.from_dict(await resp.json())

    async def set_weaning(self,
                          weaning: bool) -> Baby:
        """Updates weaning setting and returns baby-related information"""
        request_payload = {
            'settings': {
                'weaning': weaning
            }
        }

        async with self.auth.patch(SNOO_BABY_ENDPOINT, json=request_payload) as resp:
            assert resp.status == 200
            return Baby.from_dict(await resp.json())



"""PySnoo setup script."""
from setuptools import setup

_VERSION = '0.1.2'


def readme():
    """Pipe README.rst"""
    with open('README.md') as desc:
        return desc.read()


setup(
    name='pysnoo',
    packages=['pysnoo'],
    version=_VERSION,
    description='A Python library and CLI to communicate with'
                ' the Snoo Smart Baby Sleeper and Bassinet from Happiest Baby',
    long_description=readme(),
    long_description_content_type='text/markdown',
    author='Martin Riedel',
    author_email='web@riedel-it.de',
    url='https://github.com/rado0x54/pysnoo',
    license='MIT',
    python_requires='>=3.7, <4',
    include_package_data=True,
    install_requires=['oauthlib', 'aiohttp', 'pubnub>=5.0.0'],
    test_suite='tests',
    scripts=['scripts/snoo'],
    keywords=[
        'baby',
        'snoo',
        'home automation',
        ],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Other Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Topic :: Home Automation',
        'Topic :: Software Development :: Libraries :: Python Modules'
        ],
)



"""PySnoo OAuth Session."""

import json
from typing import Callable

from oauthlib.oauth2 import LegacyApplicationClient

from .const import (OAUTH_CLIENT_ID,
                    OAUTH_TOKEN_REFRESH_ENDPOINT,
                    OAUTH_LOGIN_ENDPOINT,
                    BASE_HEADERS)
from .oauth2_session import OAuth2Session


class SnooAuthSession(OAuth2Session):
    """Snoo-specific OAuth2 Session Object"""

    def __init__(
            self,
            token: dict = None,
            token_updater: Callable[[dict], None] = None) -> None:
        """Construct a new OAuth 2 client session."""

        # From Const
        super().__init__(
            client=LegacyApplicationClient(client_id=OAUTH_CLIENT_ID),
            auto_refresh_url=OAUTH_TOKEN_REFRESH_ENDPOINT,
            auto_refresh_kwargs=None,
            scope=None,
            redirect_uri=None,
            token=token,
            state=None,
            token_updater=token_updater,
            headers=BASE_HEADERS)

    async def fetch_token(self, username: str, password: str):  # pylint: disable=arguments-differ
        # Note, Snoo OAuth API is not 100% RFC 6749 compliant. (Wrong Content-Type)
        headers = {
            'Accept': 'application/json',
            'Content-Type': 'application/json;charset=UTF-8',
        }
        return await super().fetch_token(OAUTH_LOGIN_ENDPOINT, code=None, authorization_response=None,
                                         body='', auth=None, username=username, password=password, method='POST',
                                         timeout=None, headers=headers, verify_ssl=True,
                                         post_payload_modifier=json.dumps)

    async def refresh_token(self, token_url: str, **kwargs):  # pylint: disable=arguments-differ
        # Note, Snoo OAuth API is not 100% RFC 6749 compliant. (Wrong Content-Type)
        headers = {
            'Accept': 'application/json',
            'Content-Type': 'application/json;charset=UTF-8',
        }
        return await super().refresh_token(token_url, headers=headers, post_payload_modifier=json.dumps, **kwargs)
